#include <LEDA/polygon.h>
#include <LEDA/gen_polygon.h>
#include <LEDA/window.h>
#include <LEDA/geowin.h>

using namespace leda;
typedef enum {Pin, Qin, Unknown} tInFlag;   


class Intersection {
   public:
      list<point> polygon_intersection(const polygon& X, const polygon& Y);
      bool LeftOn(point x, point y, point z);
      tInFlag InOut(tInFlag inflag, bool aHB);
};


bool Intersection::LeftOn(point x, point y, point z) //LEFT-ON-TEST
{
   segment test(x, y);
   int o = orientation(test, z);
   if (o == 1) return true;
   else return false;
}


tInFlag Intersection::InOut(tInFlag inflag, bool aHB) //IN-OUT-TEST
{
   if (aHB) return Pin;
   else return Qin;   
}


list<point> Intersection::polygon_intersection(const polygon& X, const polygon& Y)
{
   //ensure a counterclockwise orientation
   polygon P, Q;
   if (X.orientation() == -1) P = X.complement(); else P = X;
   if (Y.orientation() == -1) Q = Y.complement(); else Q = Y;

   //save segments of both lists 
   segment s;
   list<segment> segments_p1;
   forall_segments(s,P) segments_p1.push_back(s);
   list<segment>::iterator A = segments_p1.begin();
   list<segment> segments_p2;
   forall_segments(s,Q) segments_p2.push_back(s);
   list<segment>::iterator B = segments_p2.begin();

   //declaration of variables and initialize them
   point a, b;               //endpoint of segment A and B respectively
   point a1, b1;             //startpoint of segment A and B respectively
   list<point> poly_result; //result list
   double cross;             //AxB cross product
   bool bHA, aHB;            //b in H(A), a in H(B)
   point p;                  //point of intersection
   tInFlag inflag = Unknown; //{Pin,Qin,Unknown) which polygon is known to be inside
   int aadv = 0, badv = 0;   //advances on a & b indices
   int n = P.size();         //size of polygon P
   int m = Q.size();         //size of polygon Q
   vector A_vector;          //saves A converted into a vector
   vector B_vector;          //saves B converted into a vector
    
   while ((aadv < n) || (badv < m)) {
      //no intersection or one contains the other --> leave the loop
      if (((aadv == n) || (badv == m)) && (inflag == Unknown)) break;
      
      cout << "\n\n\n##### WHILE-LOOP --> " << "aadv=" << aadv << ", badv=" << badv ;
      a1 = (*A).start();
      a  = (*A).end();
      b1 = (*B).start();
      b  = (*B).end();
      cout << "\nsegment(P)    = " << (*A);
      cout << "\nsegment(Q)    = " << (*B);
      
      //convert segments into vectors
      A_vector = (*A).to_vector();
      B_vector = (*B).to_vector();
      
      //calculate cross product
      cout << "\ncross product = ";
      cross = (A_vector.xcoord() * B_vector.ycoord()) - (A_vector.ycoord() * B_vector.xcoord());
      cout << cross ;
      
      //orientation test
      bHA = LeftOn(a1, a, b);
      aHB = LeftOn(b1, b, a);
        
      //intersection of current segments?
      if ((*A).intersection((*B), p))
      {
         if (inflag == Unknown)
         {  cout << "\nA and B have been advanced on a possible intersection!" << "\n-->  aadv=0, badv=0";
            aadv = 0; badv = 0;
         }
         cout << "\nRESULT POINT  = " << p << " (intersection)";
         poly_result.push_back(p);
         inflag = InOut (inflag, aHB);
      }
      
      //advance rules
      if (cross >= 0)
      {
         if (bHA) { if (inflag==Pin) { poly_result.push_back(a); cout << "\nRESULT POINT  = " << a << " (Pin)"; }
                    aadv++;
                    ++A;
                    if (A == segments_p1.end()) A = segments_p1.begin();
                    cout << "\nUsed rule     = AxB>=0 and b in H(A) --> Advance A!";
                  }
         else     { if (inflag==Qin) { poly_result.push_back(b); cout << "\nRESULT POINT  = " << b << " (Qin)"; }
                    badv++;
                    ++B;
                    if (B == segments_p2.end()) B = segments_p2.begin();
                    cout << "\nAUsed rule     = AxB>=0 and b not in H(A) --> Advance B!";
                  }
      }
      else 
      {
         if (aHB) { if (inflag==Qin) {poly_result.push_back(b); cout << "\nRESULT POINT  = " << b << " (Qin)"; }
                    badv++;
                    ++B;
                    if (B == segments_p2.end()) B = segments_p2.begin();
                    cout << "\nUsed rule     = AxB<0 and a in H(B) --> Advance B!";
                  }
         else     { if (inflag==Pin) { poly_result.push_back(a); cout << "\nRESULT POINT  = " << a << " (Pin)"; }
                    aadv++;
                    ++A; 
                    if (A == segments_p1.end()) A = segments_p1.begin();
                    cout << "\nUsed rule     = AxB<0 and b not in H(B) --> Advance A!";
                  }
      }         
   }//while  

   //deal with special cases
   if (((aadv == n) || (badv == m)) && (inflag == Unknown)) {
      Ray_crossings exception;
      bool intersection1 = false, intersection2 = false;
      point exc;
      forall_vertices(exc, P) //contains polygon Q polygon P?
      {
         intersection1 = exception.ray_crossings(Q,exc);
         if (intersection1 == false) break;
      }
      forall_vertices(exc, Q) //contains polygon P polyogn Q?
      {
         intersection2 = exception.ray_crossings(P,exc);
         if (intersection2 == false) break;
      }
      if (intersection1==false && intersection2==false) cout << "\nThere isn't any intersection!";
      if (intersection1==true || intersection2==true)
      {   if (intersection1 == true) {forall_vertices(exc,P) poly_result.push_back(exc); cout << "\n\nQ contains P!"; }
          if (intersection2 == true) {forall_vertices(exc,Q) poly_result.push_back(exc); cout << "\n\nP contains Q!"; } 
      }     
   }

   cout << "\n\n\n\n$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$";
   cout << "\n$      VERTICES OF RESULT POLYGON        $";
   cout << "\n$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n";
   polygon C(poly_result);
   point x;
   forall_vertices(x,C) cout << "\n\t" << x ;
   cout << "\n\nNumber of  vertices:\t" << poly_result.size();
   
   return poly_result;
}


